@@ -57,6 +57,7 @@
 
 #include "gdm-session-record.h"
 #include "gdm-session-worker-job.h"
+#include "gdm-common.h"
 
 #define GDM_SESSION_DBUS_PATH         "/org/gnome/DisplayManager/Session"
 #define GDM_SESSION_DBUS_INTERFACE    "org.gnome.DisplayManager.Session"
@@ -73,8 +74,6 @@ struct _GdmSessionDirectPrivate
         char                *saved_session;
         char                *selected_language;
         char                *saved_language;
-        char                *selected_layout;
-        char                *saved_layout;
         char                *selected_user;
         char                *user_x11_authority_file;
 
@@ -94,6 +93,8 @@ struct _GdmSessionDirectPrivate
         char                *display_x11_authority_file;
         gboolean             display_is_local;
 
+        char                *fallback_session_name;
+
         DBusServer          *server;
         char                *server_address;
         GHashTable          *environment;
@@ -493,7 +494,6 @@ get_session_command_for_file (const char *file,
 {
         GKeyFile   *key_file;
         GError     *error;
-        char       *full_path;
         char       *exec;
         gboolean    ret;
         gboolean    res;
@@ -509,11 +509,10 @@ get_session_command_for_file (const char *file,
         g_debug ("GdmSessionDirect: looking for session file '%s'", file);
 
         error = NULL;
-        full_path = NULL;
         res = g_key_file_load_from_dirs (key_file,
                                          file,
                                          get_system_session_dirs (),
-                                         &full_path,
+                                         NULL,
                                          G_KEY_FILE_NONE,
                                          &error);
         if (! res) {
@@ -583,11 +582,22 @@ get_session_command_for_name (const char *name,
         char    *filename;
 
         filename = g_strdup_printf ("%s.desktop", name);
-
-        command = NULL;
         res = get_session_command_for_file (filename, command);
         g_free (filename);
 
+        /*
+         * The GDM Xsession script honors "custom" as a valid session.  If the
+         * session is one of these, no file is needed, then just run the
+         * command as "custom".
+         */
+        if (!res && strcmp (name, GDM_CUSTOM_SESSION) == 0) {
+                g_debug ("No custom desktop file, but accepting it anyway.");
+                if (command != NULL) {
+                        *command = g_strdup (GDM_CUSTOM_SESSION);
+                }
+                res = TRUE;
+        }
+
         return res;
 }
 
@@ -601,64 +611,35 @@ get_default_language_name (GdmSessionDirect *session)
     return setlocale (LC_MESSAGES, NULL);
 }
 
-static char *
-get_system_default_layout (GdmSessionDirect *session)
-{
-    char *result = NULL;
-#ifdef HAVE_LIBXKLAVIER
-    static XklEngine *engine = NULL;
-    
-    if (engine == NULL) {
-	    Display *display = XOpenDisplay (session->priv->display_name);
-	    if (display != NULL) {
-		    engine = xkl_engine_get_instance (display);
-	    }
-	    /* do NOT call XCloseDisplay (display) here;
-	     * xkl_engine_get_instance() is a singleton which saves the display */
-    }
-    
-    if (engine != NULL) {
-            XklConfigRec *config = xkl_config_rec_new ();
-            if (xkl_config_rec_get_from_server (config, engine) && config->layouts && config->layouts[0]) {
-                    if (config->variants && config->variants[0] && config->variants[0][0])
-                            result = g_strdup_printf("%s\t%s", config->layouts[0], config->variants[0]);
-                    else
-                            result = g_strdup (config->layouts[0]);
-            }
-            g_object_unref (config);
-    }
-#endif
-
-    if (!result)
-        result = g_strdup ("us");    
-    return result;
-}
-
 static const char *
-get_default_layout_name (GdmSessionDirect *session)
-{
-    if (!session->priv->saved_layout) {
-        session->priv->saved_layout = get_system_default_layout (session);
-    }
-
-    return session->priv->saved_layout;
-}
-
-static char *
-get_fallback_session_name (void)
+get_fallback_session_name (GdmSessionDirect *session_direct)
 {
         const char  **search_dirs;
         int           i;
         char         *name;
+        GSequence    *sessions;
+        GSequenceIter *session;
+
+        if (session_direct->priv->fallback_session_name != NULL) {
+                /* verify that the cached version still exists */
+                if (get_session_command_for_name (session_direct->priv->fallback_session_name, NULL)) {
+                        goto out;
+                }
+        }
 
         name = g_strdup ("gnome");
         if (get_session_command_for_name (name, NULL)) {
-                return name;
+                g_free (session_direct->priv->fallback_session_name);
+                session_direct->priv->fallback_session_name = name;
+                goto out;
         }
+        g_free (name);
+
+        sessions = g_sequence_new (g_free);
 
         search_dirs = get_system_session_dirs ();
         for (i = 0; search_dirs[i] != NULL; i++) {
-                GDir *dir;
+                GDir       *dir;
                 const char *base_name;
 
                 dir = g_dir_open (search_dirs[i], 0, NULL);
@@ -679,19 +660,38 @@ get_fallback_session_name (void)
                         }
 
                         if (get_session_command_for_file (base_name, NULL)) {
-                                g_free (name);
 
-                                name = g_strndup (base_name,
-                                                  strlen (base_name) -
-                                                  strlen (".desktop"));
-                                break;
+                                g_sequence_insert_sorted (sessions, g_strdup (base_name), (GCompareDataFunc) g_strcmp0, NULL);
                         }
                 } while (base_name != NULL);
 
                 g_dir_close (dir);
         }
 
-        return name;
+        name = NULL;
+        session = g_sequence_get_begin_iter (sessions);
+        do {
+               if (g_sequence_get (session)) {
+                       char *base_name;
+
+                       g_free (name);
+                       base_name = g_sequence_get (session);
+                       name = g_strndup (base_name,
+                                         strlen (base_name) -
+                                         strlen (".desktop"));
+
+                       break;
+               }
+               session = g_sequence_iter_next (session);
+        } while (!g_sequence_iter_is_end (session));
+
+        g_free (session_direct->priv->fallback_session_name);
+        session_direct->priv->fallback_session_name = name;
+
+        g_sequence_free (sessions);
+
+ out:
+        return session_direct->priv->fallback_session_name;
 }
 
 static const char *
@@ -701,7 +701,7 @@ get_default_session_name (GdmSessionDirect *session)
                 return session->priv->saved_session;
         }
 
-        return get_fallback_session_name ();
+        return get_fallback_session_name (session);
 }
 
 static void
@@ -709,8 +709,6 @@ gdm_session_direct_defaults_changed (GdmSessionDirect *session)
 {
         _gdm_session_default_language_name_changed (GDM_SESSION (session),
                                                     get_default_language_name (session));
-        _gdm_session_default_layout_name_changed (GDM_SESSION (session),
-                                                  get_default_layout_name (session));
         _gdm_session_default_session_name_changed (GDM_SESSION (session),
                                                    get_default_session_name (session));
 }
@@ -731,9 +729,6 @@ gdm_session_direct_select_user (GdmSession *session,
 
         g_free (impl->priv->saved_language);
         impl->priv->saved_language = NULL;
-
-        g_free (impl->priv->saved_layout);
-        impl->priv->saved_layout = NULL;
 }
 
 static DBusHandlerResult
@@ -935,6 +930,58 @@ gdm_session_direct_handle_problem (GdmSessionDirect *session,
         return DBUS_HANDLER_RESULT_HANDLED;
 }
 
+static DBusHandlerResult
+gdm_session_direct_handle_session_opened (GdmSessionDirect *session,
+                                          DBusConnection   *connection,
+                                          DBusMessage      *message)
+{
+        DBusMessage *reply;
+        DBusError    error;
+
+        g_debug ("GdmSessionDirect: Handling SessionOpened");
+
+        dbus_error_init (&error);
+        if (! dbus_message_get_args (message, &error, DBUS_TYPE_INVALID)) {
+                g_warning ("ERROR: %s", error.message);
+        }
+
+        g_debug ("GdmSessionDirect: Emitting 'session-opened' signal");
+
+        _gdm_session_session_opened (GDM_SESSION (session));
+
+        reply = dbus_message_new_method_return (message);
+        dbus_connection_send (connection, reply, NULL);
+        dbus_message_unref (reply);
+
+        return DBUS_HANDLER_RESULT_HANDLED;
+}
+
+static DBusHandlerResult
+gdm_session_direct_handle_open_failed (GdmSessionDirect *session,
+                                       DBusConnection   *connection,
+                                       DBusMessage      *message)
+{
+        DBusMessage *reply;
+        DBusError    error;
+        const char  *text;
+
+        dbus_error_init (&error);
+        if (! dbus_message_get_args (message, &error,
+                                     DBUS_TYPE_STRING, &text,
+                                     DBUS_TYPE_INVALID)) {
+                g_warning ("ERROR: %s", error.message);
+        }
+
+        reply = dbus_message_new_method_return (message);
+        dbus_connection_send (connection, reply, NULL);
+        dbus_message_unref (reply);
+
+        g_debug ("GdmSessionDirect: Emitting 'session-open-failed' signal");
+        _gdm_session_session_open_failed (GDM_SESSION (session), text);
+
+        return DBUS_HANDLER_RESULT_HANDLED;
+}
+
 static DBusHandlerResult
 gdm_session_direct_handle_session_started (GdmSessionDirect *session,
                                            DBusConnection   *connection,
@@ -1087,39 +1134,6 @@ gdm_session_direct_handle_saved_language_name_read (GdmSessionDirect *session,
         return DBUS_HANDLER_RESULT_HANDLED;
 }
 
-static DBusHandlerResult
-gdm_session_direct_handle_saved_layout_name_read (GdmSessionDirect *session,
-                                                  DBusConnection   *connection,
-                                                  DBusMessage      *message)
-{
-        DBusMessage *reply;
-        DBusError    error;
-        const char  *layout_name;
-
-        dbus_error_init (&error);
-        if (! dbus_message_get_args (message, &error,
-                                     DBUS_TYPE_STRING, &layout_name,
-                                     DBUS_TYPE_INVALID)) {
-                g_warning ("ERROR: %s", error.message);
-        }
-
-        reply = dbus_message_new_method_return (message);
-        dbus_connection_send (connection, reply, NULL);
-        dbus_message_unref (reply);
-
-        if (strcmp (layout_name,
-                    get_default_layout_name (session)) != 0) {
-                g_free (session->priv->saved_layout);
-                session->priv->saved_layout = g_strdup (layout_name);
-
-                _gdm_session_default_layout_name_changed (GDM_SESSION (session),
-                                                          layout_name);
-        }
-
-
-        return DBUS_HANDLER_RESULT_HANDLED;
-}
-
 static DBusHandlerResult
 gdm_session_direct_handle_saved_session_name_read (GdmSessionDirect *session,
                                                    DBusConnection   *connection,
@@ -1199,6 +1213,10 @@ session_worker_message (DBusConnection *connection,
                 return gdm_session_direct_handle_accreditation_failed (session, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "UsernameChanged")) {
                 return gdm_session_direct_handle_username_changed (session, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SessionOpened")) {
+                return gdm_session_direct_handle_session_opened (session, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "OpenFailed")) {
+                return gdm_session_direct_handle_open_failed (session, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SessionStarted")) {
                 return gdm_session_direct_handle_session_started (session, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "StartFailed")) {
@@ -1209,8 +1227,6 @@ session_worker_message (DBusConnection *connection,
                 return gdm_session_direct_handle_session_died (session, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SavedLanguageNameRead")) {
                 return gdm_session_direct_handle_saved_language_name_read (session, connection, message);
-        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SavedLayoutNameRead")) {
-                return gdm_session_direct_handle_saved_layout_name_read (session, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SavedSessionNameRead")) {
                 return gdm_session_direct_handle_saved_session_name_read (session, connection, message);
         }
@@ -1492,8 +1508,8 @@ handle_connection (DBusServer      *server,
                                                       &vtable,
                                                       session);
 
-                g_debug ("GdmSessionDirect: Emitting opened signal");
-                _gdm_session_opened (GDM_SESSION (session));
+                g_debug ("GdmSessionDirect: Emitting conversation-started signal");
+                _gdm_session_conversation_started (GDM_SESSION (session));
         }
 }
 
@@ -1656,13 +1672,13 @@ stop_worker (GdmSessionDirect *session)
 }
 
 static void
-gdm_session_direct_open (GdmSession *session)
+gdm_session_direct_start_conversation (GdmSession *session)
 {
         GdmSessionDirect *impl = GDM_SESSION_DIRECT (session);
 
         g_return_if_fail (session != NULL);
 
-        g_debug ("GdmSessionDirect: Opening session");
+        g_debug ("GdmSessionDirect: Starting conversation");
 
         start_worker (impl);
 }
@@ -1897,16 +1913,6 @@ get_language_name (GdmSessionDirect *session)
         return get_default_language_name (session);
 }
 
-static const char *
-get_layout_name (GdmSessionDirect *session)
-{
-        if (session->priv->selected_layout != NULL) {
-                return session->priv->selected_layout;
-        }
-
-        return get_default_layout_name (session);
-}
-
 static const char *
 get_session_name (GdmSessionDirect *session)
 {
@@ -1922,20 +1928,18 @@ get_session_name (GdmSessionDirect *session)
 static char *
 get_session_command (GdmSessionDirect *session)
 {
-        gboolean res;
-        char    *command;
-        char    *filename;
+        gboolean    res;
+        char       *command;
+        const char *session_name;
 
-        filename = g_strdup_printf ("%s.desktop", get_session_name (session));
+        session_name = get_session_name (session);
 
         command = NULL;
-        res = get_session_command_for_file (filename, &command);
+        res = get_session_command_for_name (session_name, &command);
         if (! res) {
-                g_critical ("Cannot read specified session file: %s", filename);
-                g_free (filename);
+                g_critical ("Cannot find a command for specified session: %s", session_name);
                 exit (1);
         }
-        g_free (filename);
 
         return command;
 }
@@ -1972,13 +1976,6 @@ setup_session_environment (GdmSessionDirect *session)
                                                      "GDM_LANG",
                                                      get_language_name (session));
 
-        if (strcmp (get_layout_name (session),
-                    get_default_layout_name (session)) == 0) {
-                gdm_session_direct_set_environment_variable (session,
-                                                             "GDM_KEYBOARD_LAYOUT",
-                                                             get_layout_name (session));
-        }
-
         gdm_session_direct_set_environment_variable (session,
                                                      "DISPLAY",
                                                      session->priv->display_name);
@@ -1989,6 +1986,13 @@ setup_session_environment (GdmSessionDirect *session)
                                                              session->priv->user_x11_authority_file);
         }
 
+        if (g_getenv ("WINDOWPATH") != NULL) {
+                gdm_session_direct_set_environment_variable (session,
+                                                             "WINDOWPATH",
+                                                             g_getenv ("WINDOWPATH"));
+        }
+
+
         /* FIXME: We do this here and in the session worker.  We should consolidate
          * somehow.
          */
@@ -2000,6 +2004,16 @@ setup_session_environment (GdmSessionDirect *session)
 
 }
 
+static void
+gdm_session_direct_open_session (GdmSession *session)
+{
+        GdmSessionDirect *impl = GDM_SESSION_DIRECT (session);
+
+        g_return_if_fail (session != NULL);
+
+        send_dbus_void_signal (impl, "OpenSession");
+}
+
 static void
 gdm_session_direct_start_session (GdmSession *session)
 {
@@ -2063,12 +2077,6 @@ gdm_session_direct_close (GdmSession *session)
         g_free (impl->priv->saved_language);
         impl->priv->saved_language = NULL;
 
-        g_free (impl->priv->selected_layout);
-        impl->priv->selected_layout = NULL;
-
-        g_free (impl->priv->saved_layout);
-        impl->priv->saved_layout = NULL;
-
         g_free (impl->priv->user_x11_authority_file);
         impl->priv->user_x11_authority_file = NULL;
 
@@ -2107,6 +2115,14 @@ gdm_session_direct_get_username (GdmSessionDirect *session)
         return g_strdup (session->priv->selected_user);
 }
 
+char *
+gdm_session_direct_get_display_device (GdmSessionDirect *session)
+{
+        g_return_val_if_fail (session != NULL, NULL);
+
+        return g_strdup (session->priv->display_device);
+}
+
 gboolean
 gdm_session_direct_bypasses_xsession (GdmSessionDirect *session_direct)
 {
@@ -2115,7 +2131,6 @@ gdm_session_direct_bypasses_xsession (GdmSessionDirect *session_direct)
         gboolean    res;
         gboolean    bypasses_xsession = FALSE;
         char       *filename;
-        char       *full_path;
 
         g_return_val_if_fail (session_direct != NULL, FALSE);
         g_return_val_if_fail (GDM_IS_SESSION_DIRECT (session_direct), FALSE);
@@ -2127,7 +2142,7 @@ gdm_session_direct_bypasses_xsession (GdmSessionDirect *session_direct)
         res = g_key_file_load_from_dirs (key_file,
                                          filename,
                                          get_system_session_dirs (),
-                                         &full_path,
+                                         NULL,
                                          G_KEY_FILE_NONE,
                                          &error);
         if (! res) {
@@ -2192,24 +2207,6 @@ gdm_session_direct_select_language (GdmSession *session,
                                  get_language_name (impl));
 }
 
-static void
-gdm_session_direct_select_layout (GdmSession *session,
-                                  const char *text)
-{
-        GdmSessionDirect *impl = GDM_SESSION_DIRECT (session);
-
-        g_free (impl->priv->selected_layout);
-
-        if (strcmp (text, "__previous") == 0) {
-                impl->priv->selected_layout = NULL;
-        } else {
-                impl->priv->selected_layout = g_strdup (text);
-        }
-
-        send_dbus_string_signal (impl, "SetLayoutName",
-                                 get_layout_name (impl));
-}
-
 static void
 _gdm_session_direct_set_display_id (GdmSessionDirect *session,
                                     const char       *id)
@@ -2402,8 +2399,8 @@ gdm_session_direct_finalize (GObject *object)
         g_free (session->priv->saved_session);
         g_free (session->priv->selected_language);
         g_free (session->priv->saved_language);
-        g_free (session->priv->selected_layout);
-        g_free (session->priv->saved_layout);
+
+        g_free (session->priv->fallback_session_name);
 
         parent_class = G_OBJECT_CLASS (gdm_session_direct_parent_class);
 
@@ -2466,12 +2463,13 @@ gdm_session_direct_constructor (GType                  type,
 static void
 gdm_session_iface_init (GdmSessionIface *iface)
 {
-        iface->open = gdm_session_direct_open;
+        iface->start_conversation = gdm_session_direct_start_conversation;
         iface->setup = gdm_session_direct_setup;
         iface->setup_for_user = gdm_session_direct_setup_for_user;
         iface->authenticate = gdm_session_direct_authenticate;
         iface->authorize = gdm_session_direct_authorize;
         iface->accredit = gdm_session_direct_accredit;
+        iface->open_session = gdm_session_direct_open_session;
         iface->close = gdm_session_direct_close;
 
         iface->cancel = gdm_session_direct_cancel;
@@ -2479,7 +2477,6 @@ gdm_session_iface_init (GdmSessionIface *iface)
         iface->answer_query = gdm_session_direct_answer_query;
         iface->select_session = gdm_session_direct_select_session;
         iface->select_language = gdm_session_direct_select_language;
-        iface->select_layout = gdm_session_direct_select_layout;
         iface->select_user = gdm_session_direct_select_user;
 }
 
